from onliner_parser.managers.input_manager import InputManager
# from onliner_parser.managers.load_manager import LoadManager - don't working at this time
from onliner_parser.managers.save_manager import SaveManager



from typing import Optional

from pydantic import BaseModel


class BasePrice(BaseModel):
    amount: float = None


class MinPricesMedian(BasePrice):
    pass


class PriceMin(BasePrice):
    pass


class PriceMax(BasePrice):
    pass


class Sale(BaseModel):
    is_on_sale: bool = None
    discount: int = None
    min_prices_median: MinPricesMedian = None


class Offers(BaseModel):
    count: int = None


class Prices(BaseModel):
    price_min: PriceMin = None
    price_max: PriceMax = None
    offers: Offers = None


class ProductReviews(BaseModel):
    rating: int = None
    count: int = None
    html_url: str = None


class Page(BaseModel):
    limit: int = None
    items: int = None
    current: int = None
    last: int = None


class PriceHistoryLog(BaseModel):
    date: Optional[str] = None
    price: Optional[float] = None


class ChartData(BaseModel):
    items: list[PriceHistoryLog]

    def get_items_tuple(self) -> tuple:
        items = []
        for item in self.items:
            data = item.date
            price = item.price
            items.append((data, price))

        return tuple(items)


class Product(BaseModel):
    id: int = None
    key: str = None
    full_name: str = None
    name_prefix: str = None
    status: str = None
    description: str = None
    html_url: str = None
    reviews: ProductReviews = None
    prices: Prices = None
    sale: Sale = None

    price_history: tuple = None
    item_spec: dict = None

    def to_dict(self) -> dict:
        dict_attrs = {
            'id': self.id,
            'key': self.key,
            'full_name': self.full_name,
            'name_prefix': self.name_prefix,
            'description': self.description,
            'html_url': self.html_url,
            'reviews_rating': self.reviews.rating,
            'reviews_html_url': self.reviews.html_url,
            'reviews_count': self.reviews.count,
            'sale_is_on_sale': self.sale.is_on_sale,
            'sale_discount': self.sale.discount,
            'sale_min_prices_median_amount': self.sale.min_prices_median.amount,
            'item_spec': self.item_spec,
        }
        if self.prices:
            dict_attrs.update(
                {
                    'prices_price_min_amount': self.prices.price_min.amount,
                    'prices_price_max_amount': self.prices.price_max.amount,
                    'prices_offers_count': self.prices.offers.count,
                    'price_history': self.price_history,
                }
            )

        return dict_attrs

    @staticmethod
    def get_fields():
        return (
            'id',
            'key',
            'full_name',
            'name_prefix',
            'description',
            'html_url',
            'reviews_rating',
            'reviews_html_url',
            'reviews_count',
            'prices_price_min_amount',
            'prices_price_max_amount',
            'prices_offers_count',
            'sale_is_on_sale',
            'sale_discount',
            'sale_min_prices_median_amount',
            'price_history',
            'item_spec',
        )


class BaseJSONResponse(BaseModel):
    products: list[Product]
    total: int
    page: Page
    total_ungrouped: int

    def get_products(self) -> list[Product]:
        return [
            product for product in self.products
            if product.status == 'active'
        ]

    def get_last_page(self) -> int:
        return self.page.last


class BasePriceHistoryJSONResponse(BaseModel):
    chart_data: ChartData = None

    def get_items(self) -> tuple:
        return self.chart_data.get_items_tuple()



from onliner_parser.utils.data_transformation import DataTransformer
from onliner_parser.utils.decorators import exec_time
from onliner_parser.utils.font_settings import Font
from onliner_parser.utils.settings import Settings



from enum import Enum


class Font(str, Enum):
    NORMAL = '\033[0m'
    BOLD = '\033[1m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'

    WARN = f'{RED}{BOLD}[WARNING]{NORMAL}{RED}'
    ERROR = f'{RED}{BOLD}[ERROR]{NORMAL}{RED}'
    INFO = f'{GREEN}{BOLD}[INFO]{NORMAL}{GREEN}'



import pandas as pd

from onliner_parser.models import Product


class DataTransformer:
    @staticmethod
    def __get_price_history_df(products_data: list[Product]) -> pd.DataFrame:
        """
        Transforms initial tuple[tuple[str, float]] structure to dataframe str - float
        Parameters
        ----------
        products_data: list[Product]

        Returns
        -------
        desc_df: pd.Dataframe()
            df with month - price columns

        """
        p_h_data: list[list] = []
        columns = [tuple_[0] + "_price" for tuple_ in products_data[0].price_history[:-1]]
        columns.insert(0, "id")
        for product in products_data:
            price_history = list(map(list, product.price_history[:-1]))

            price_history = [list_[1] for list_ in price_history]
            price_history.insert(0, product.id)

            p_h_data.append(price_history)
        return pd.DataFrame(p_h_data, columns=columns)

    @staticmethod
    def __get_specification_df(products_data: list[Product]) -> pd.DataFrame:
        """
        Transforms initial dict[str, [dict[str,str] | str]] structure to dataframe of strs
        Parameters
        ----------
        products_data: list[Product]

        Returns
        -------
        desc_df: pd.Dataframe()
            df with columns that descript product

        """
        spec_df = pd.DataFrame()
        for prod in products_data:
            df_ = pd.DataFrame.from_dict({j: prod.item_spec[i][j]
                                          for i in prod.item_spec.keys()
                                          for j in prod.item_spec[i].keys()},
                                         orient="index")
            df_ = df_.transpose()
            df_["id"] = prod.id
            spec_df = pd.concat([spec_df, df_])
        spec_df.reset_index(inplace=True, drop=True)
        return spec_df

    def transform_dp_fields(self, full_df: pd.DataFrame(), products_data: list[Product], settings) -> pd.DataFrame():
        """
        Transforms deep parsed fields
        Parameters
        ----------
        full_df: pd.DataFrame()
        products_data: list[Product]
        settings: Settings
            class of bool vars(parse or not)

        Returns
        -------
        full_df: pd.Dataframe()
            transformed df

        """
        full_df.drop(["price_history", "item_spec"], axis=1, inplace=True)

        # transform fields from deep parsing
        if settings.deep_parse_status:
            if settings.parse_history:
                p_h_df = self.__get_price_history_df(products_data)
                full_df = full_df.merge(p_h_df, on="id")
            if settings.parse_spec:
                spec_df = self.__get_specification_df(products_data)
                full_df = full_df.merge(spec_df, on="id")
        return full_df



import time
from functools import wraps

from onliner_parser.utils.font_settings import Font


def exec_time(start_text='Starting parsing...', end_text='Parsing completed!'):
    def wrapper(func):
        @wraps(func)
        def wrapped_wrapper(*args, **kwargs):
            start = time.time()
            print(f'{Font.INFO} {start_text}')
            func(*args, **kwargs)
            executing_time = time.strftime("%H:%M:%S", time.gmtime(round(time.time() - start, 2)))
            print(f'\n{Font.INFO} {end_text} Execution time: {executing_time}{Font.NORMAL}')
        return wrapped_wrapper
    return wrapper



import pandas as pd
import numpy as np
import plotly.graph_objects as go


class DataProcManager:
    def __init__(self, file=None, ) -> None:
        self.file = file
        self.data = pd.read_csv(file)

    @staticmethod
    def __get_products_on_review_rate() -> list[int]:
        """
        Gets the best products on review rate
        Returns
        -------
        product_ids: list[int]

        """

        return [42342]

    @staticmethod
    def __get_products_on_price_review_rate() -> list[int]:
        """
        Gets the best products on price review rate
        Returns
        -------
        product_ids: list[int]

        """

        return [42342]

    @staticmethod
    def __get_shop_offers_on_price_review_rate() -> list[int]:
        """
        Gets the best shop offers on price review rate
        Returns
        -------
        shop_ids: list[int]

        """

        return [42342]

    @staticmethod
    def __get_roducts_price_history_plots() -> list[bytes]:
        """
        Gets the items price history plots
        Returns
        -------
        img_bytes: list[bytes]
        """
        np.random.seed(1)

        N = 100
        x = np.random.rand(N)
        y = np.random.rand(N)
        colors = np.random.rand(N)
        sz = np.random.rand(N) * 30

        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x=x,
            y=y,
            mode="markers",
            marker=go.scatter.Marker(
                size=sz,
                color=colors,
                opacity=0.6,
                colorscale="Viridis"
            )
        ))
        img_bytes = fig.to_image(format="png")
        return [img_bytes]



import asyncio
import time
from random import uniform

import bs4
from aiohttp import ClientSession
from progress.bar import IncrementalBar
from requests import Response, Session
from requests.exceptions import ConnectionError

from onliner_parser.models import (BaseJSONResponse,
                                   BasePriceHistoryJSONResponse, Product)
from onliner_parser.utils import Font, Settings, exec_time


class CatalogParser:
    __session: Session
    SETTINGS: Settings

    __url: str = 'https://catalog.onliner.by/sdapi/catalog.api/search/'
    __headers: dict = {
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                      'Chrome/106.0.0.0 Safari/537.36 Edg/106.0.1370.42',
        'x-requested-with': 'XMLHttpRequest',
        'sec-fetch-site': 'same-origin',
        'accept-language': 'ru,en;q=0.9,en-GB;q=0.8,en-US;q=0.7',
        'accept-encoding': 'gzip, deflate, br',
        'accept': 'application/json, text/javascript, */*; q=0.01',
    }
    __params: dict

    __base_json_response: BaseJSONResponse
    __data: list[Product] = []

    __last_page: int
    __items_count: int
    __current_item_index: int

    def __init__(self, url: str) -> None:
        category = url.split('/')[-1] + '?group=1'
        self.__url += category

        self.__session = Session()
        self.SETTINGS = Settings()

        self.__default__init__()

    def __del__(self) -> None:
        self.__session.close()

    def __default__init__(self) -> None:
        self.__params: dict = {'page': 1}
        self.__data: list[Product] = []
        self.__last_page: int = 0
        self.__items_count: int = 0
        self.__current_item_index: int = 0

    @staticmethod
    def __random_wait(start: float = 0.1, finish: float = 0.3) -> None:
        """Random waiting in the specified range"""
        sec = uniform(start, finish)
        time.sleep(sec)

    def __get_response(self, url: str = None) -> Response:
        """Get response from Site"""
        response: Response
        while True:
            try:
                if url:
                    response = self.__session.get(url)
                else:
                    response = self.__session.get(self.__url, headers=self.__headers, params=self.__params)
            except ConnectionError:
                self.__random_wait(1, 2)
            else:
                if response.status_code == 200:
                    return response
                else:
                    self.__random_wait()

    def __set_base_json_response(self, json: str) -> None:
        """
        Set BaseJSONResponse
        json: str - json response
        """
        self.__base_json_response = BaseJSONResponse.parse_raw(json)

    def __set_last_page(self, page: int) -> None:
        """
        Set last page for parse
        page: int - page number
        """
        self.__last_page = page

    def __extend_data(self, data: list[Product]) -> None:
        """Adding new data to memory"""
        self.__data.extend(data)

    def __inc_current_page(self) -> bool:
        """Increasing the current parsing page until it is equal to the last page"""
        self.__params['page'] = self.__params.get("page") + 1
        if self.__params['page'] <= self.__last_page:
            return True
        return False

    def __inc_current_item_index(self) -> bool:
        """Increasing current item index until it is equal items count"""
        if self.__current_item_index < self.get_items_count() - 1:  # -1 because list start from 0
            self.__current_item_index += 1
            return True
        self.SETTINGS.deep_parse_status = True
        return False

    @staticmethod
    def __get_price_history_link(key: str) -> str:
        return f'https://catalog.api.onliner.by/products/{key}/prices-history?period=6m'

    @staticmethod
    def __bs_parse(html: str) -> dict:
        soup = bs4.BeautifulSoup(html, 'html.parser')
        [bad_div.decompose() for bad_div in soup.find_all('div', class_='product-tip-wrapper')]
        tbodys = soup.select('table[class="product-specs__table"] tbody')
        specs = dict()
        for tbody in tbodys:
            tr_list = tbody.find_all('tr')
            title = tr_list[0].text.strip()
            info = dict()
            for tr in tr_list[1:]:
                names = [" ".join(tag.text.split()) for tag in tr.select('td') if tag.text]
                descs = [" ".join(tag.text.split()) for tag in tr.select('td span[class="value__text"]')]
                dict_ = dict(zip(names, descs))
                info.update(dict_)
            specs.update({title: info})
        return specs

    def __get_price_history(self, key: str) -> tuple:
        url = self.__get_price_history_link(key)

        response = self.__get_response(url)

        base_price_history_json_response = BasePriceHistoryJSONResponse().parse_raw(response.text)
        return base_price_history_json_response.get_items()

    def __get_item_spec(self, url: str) -> dict:
        """Parse item spec by bs4"""
        response = self.__get_response(url)

        return self.__bs_parse(response.text)

    def __deep_parse_item(self, item: Product) -> None:
        """Deep parsing given item"""
        if self.SETTINGS.parse_history:
            item.price_history = self.__get_price_history(item.key)
        if self.SETTINGS.parse_spec:
            item.item_spec = self.__get_item_spec(item.html_url)

    async def __get_async_response(self, async_session: ClientSession, page: int):
        async with async_session.get(self.__url, params={'page': page}) as response:
            json = await response.text()
            bjr: BaseJSONResponse = BaseJSONResponse.parse_raw(json)
            self.__extend_data(bjr.get_products())

    async def __get_future_instance(self):
        tasks = []
        async with ClientSession() as async_session:
            for page in range(1, self.__last_page + 1):
                task = asyncio.ensure_future(self.__get_async_response(async_session, page=page))
                tasks.append(task)

            await asyncio.gather(*tasks)

    async def __async_deep_parse_item(self, async_session: ClientSession, item: Product):
        if self.SETTINGS.parse_spec:
            async with async_session.get(item.html_url) as response:
                html = await response.text()
                item.item_spec = self.__bs_parse(html)
        if self.SETTINGS.parse_history:
            url = self.__get_price_history_link(item.key)
            async with async_session.get(url) as response:
                json = await response.text()
                base_price_history: BasePriceHistoryJSONResponse = BasePriceHistoryJSONResponse().parse_raw(json)
                item.price_history = base_price_history.get_items()

    async def __get_future_instance_deep_parse(self):
        tasks = []
        async with ClientSession() as async_session:
            for prod in self.__data:
                task = asyncio.ensure_future(self.__async_deep_parse_item(async_session, prod))
                tasks.append(task)

            await asyncio.gather(*tasks)

    @exec_time()
    def __async_parse(self) -> None:
        """Async parsing process"""
        response: Response = self.__get_response()

        self.__set_base_json_response(response.text)
        self.__set_last_page(self.__base_json_response.get_last_page())

        self.__extend_data(self.__base_json_response.get_products())

        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(self.__get_future_instance())
        loop.run_until_complete(future)

        self.__items_count = len(self.__data)

    @exec_time(start_text='Start deep parsing...', end_text='Deep parsing completed!')
    def __async_deep_parse(self) -> None:
        """Async deep parsing"""
        if not self.__data:
            self.__async_parse()

        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(self.__get_future_instance_deep_parse())
        loop.run_until_complete(future)

        self.SETTINGS.deep_parse_status = True

    @exec_time()
    def __parse(self) -> None:
        """Parsing process"""
        response: Response = self.__get_response()

        self.__set_base_json_response(response.text)
        self.__set_last_page(self.__base_json_response.get_last_page())

        self.__extend_data(self.__base_json_response.get_products())
        # Show progress
        bar = IncrementalBar(f'{Font.YELLOW}Parsing process:{Font.NORMAL}', max=self.__last_page)
        bar.next()

        while self.__inc_current_page():
            response = self.__get_response()
            self.__set_base_json_response(response.text)
            self.__extend_data(self.__base_json_response.get_products())
            bar.next()
            self.__random_wait()

        self.__items_count = len(self.__data)

    def __parse_next(self) -> bool:
        """Parse next item in data"""
        if not self.__data:
            self.__parse()

        current_index = self.__current_item_index
        self.__deep_parse_item(self.__data[current_index])

        return self.__inc_current_item_index()

    @exec_time(start_text='Start deep parsing...', end_text='Deep parsing completed!')
    def __deep_parse(self) -> None:
        """Deep parsing process"""
        if not self.__data:
            self.__parse()

        bar = IncrementalBar(f'{Font.YELLOW}Deep parsing process:{Font.NORMAL}', max=len(self.__data))

        for item in self.__data:
            self.__deep_parse_item(item)
            bar.next()

        self.SETTINGS.deep_parse_status = True

    def parse(self) -> None:
        """Start parsing"""
        self.__parse()

    def async_parse(self) -> None:
        """Start async parsing"""
        self.__async_parse()

    def deep_parse(self) -> None:
        """Start deep parsing"""
        self.__deep_parse()

    def async_deep_parse(self) -> None:
        """Start async deep parse"""
        self.__async_deep_parse()

    def parse_next(self) -> bool:
        """Parse next item in data"""
        return self.__parse_next()

    def get_data(self) -> list[Product]:
        """Returns the parsed data"""
        if self.__data:
            return self.__data
        print(f'{Font.ERROR} Nothing to return')

    def insert_data(self, data: list[Product]) -> None:
        """Insert data into parser"""
        if not self.__data:
            self.__data = data
        else:
            print(f'{Font.ERROR} The parser already contains data')

    def get_items_count(self) -> int:
        """Return count of items"""
        if self.__items_count:
            return self.__items_count
        print(f'{Font.ERROR} Nothing to return')



import csv

from onliner_parser.models import (MinPricesMedian, Offers, PriceMax, PriceMin,
                                   Prices, Product, ProductReviews, Sale)

# NOT SUPPORTED AT THIS TIME!!! DON'T USE THIS MANAGER!!!


class LoadManager:
    __data: list[Product] = []
    __directory_name: str = 'data/'

    def load_csv(self, filename: str = 'products') -> None:
        filepath = f'{self.__directory_name}{filename}.csv'

        with open(filepath, newline='', encoding='utf-8') as file:
            reader = csv.DictReader(file)

            for row in reader:
                reviews_rating = int(row.pop('reviews_rating'))
                reviews_html_url = str(row.pop('reviews_html_url'))
                reviews_count = int(row.pop('reviews_count'))
                reviews = ProductReviews(rating=reviews_rating, count=reviews_count, html_url=reviews_html_url)

                prices_price_min_amount = row.pop('prices_price_min_amount')
                prices_price_max_amount = row.pop('prices_price_max_amount')
                price_min = PriceMin(amount=float(prices_price_min_amount) if prices_price_min_amount else None)
                price_max = PriceMax(amount=float(prices_price_max_amount) if prices_price_max_amount else None)
                prices_offers_count = row.pop('prices_offers_count')
                offers = Offers(count=int(prices_offers_count) if prices_offers_count else None)
                prices = Prices(price_min=price_min, price_max=price_max, offers=offers)

                is_on_sale = True if row.pop('sale_is_on_sale') == "True" else False
                discount = int(row.pop('sale_discount'))
                min_prices_median = MinPricesMedian(amount=float(row.pop('sale_min_prices_median_amount')))
                sale = Sale(is_on_sale=is_on_sale, discount=discount, min_prices_median=min_prices_median)

                price_history = str(row.pop('price_history'))

                item_spec = str(row.pop('item_spec'))

                product = Product.parse_obj(row)
                product.reviews = reviews
                product.prices = prices
                product.sale = sale
                product.price_history = price_history
                product.item_spec = item_spec

                self.__data.append(product)

    def get_data(self) -> list[Product]:
        return self.__data



import argparse
import os


class InputManager:
    @staticmethod
    def get_args() -> dict:
        if not os.environ.get("IS_DOCKER", False):
            arg_parser = argparse.ArgumentParser(description='Parser Catalog Onliner')

            arg_parser.add_argument(
                '-u',
                dest='url',
                type=str,
                help='URL category',
                required=True,
            )

            arg_parser.add_argument(
                '-fn',
                dest='filename',
                type=str,
                help='The name of the file to which the data will be saved (standard - products)',
                default='products',
                required=False,
            )

            arg_parser.add_argument(
                '-sf',
                dest='save_format',
                type=str,
                help='Format for saving data (csv/xlsx) (standard - csv)',
                default='csv',
                required=False,
            )

            args = arg_parser.parse_args()

            url = args.url
            filename = args.filename
            save_format = args.save_format
        else:
            url = os.environ.get("URL")
            filename = os.environ.get("FILENAME", "products")
            save_format = os.environ.get("SAVE_FORMAT", "csv")

        args = {
            'url': url,
            'filename': filename,
            'save_format': save_format,
        }

        return args



import os

import pandas as pd

from onliner_parser.models import Product
from onliner_parser.utils import DataTransformer, Font, Settings


class SaveManager:
    __settings: Settings
    __data: list[Product]
    __directory_name: str = 'data/'

    excel_formats: tuple = (
        'xlsx',
        'xlsm',
        'xlsb',
        'xlam',
        'xltx',
        'xltm',
        'xls',
        'xla',
        'xlt',
        'excel',
    )

    def __init__(self, data: list[Product], settings: Settings) -> None:
        exception = ValueError('You must pass an instance of the list[Product] and Settings')

        for prod in data:
            if not isinstance(prod, Product):
                raise exception
        self.__data = data

        if isinstance(settings, Settings):
            self.__settings = settings
        else:
            raise exception

    def __create_directory(self):
        if not os.path.exists(f'{self.__directory_name}'):
            os.makedirs(f'{self.__directory_name}')

    def set_directory_name(self, name: str) -> None:
        self.__directory_name = name

    def save(self, filename: str = 'products', save_format: str = 'csv') -> None:
        if self.__data:
            self.__create_directory()

            list_of_prods = [product.to_dict() for product in self.__data]
            df = pd.DataFrame(list_of_prods)

            data_transformer = DataTransformer()
            df = data_transformer.transform_dp_fields(df, self.__data, self.__settings)

            filepath = f'{self.__directory_name}{filename}.{save_format}'
            try:
                if save_format == 'csv':
                    df.to_csv(filepath, index=False)
                else:
                    df.to_excel(filepath, sheet_name='Products', na_rep='-', index=False)
            except BaseException as e:
                print(f'{Font.WARN} Recording error{Font.NORMAL}')
                print(f'{Font.ERROR} {e}')

            else:
                print(f'{Font.INFO} Data saved to {filename}.{save_format}{Font.NORMAL}')

        else:
            print(f'{Font.WARN} There is nothing to save{Font.NORMAL}')



from setuptools import find_packages, setup

requirements = [
    "pydantic",
    "requests",
    "progress",
    "pandas",
    "openpyxl",
    "beautifulsoup4",
    "aiohttp",
]

setup(
    name='onliner_parser',
    version='0.2.2.1',
    description='Парсер для каталога Onliner.by',
    packages=find_packages(),
    install_requires=requirements,
    author_email='aalshe38@gmail.com',
    zip_safe=False,
)
